import logging
from datetime import datetime
from typing import Iterable, Optional, Tuple

import pytz
from cachetools.func import ttl_cache

from fetchfox.apis.algorand import algonodecloud, nfdomains, randswapcom
from fetchfox.blockchains.base import Blockchain
from fetchfox.constants.blockchains import ALGORAND
from fetchfox.constants.currencies import ALGO
from fetchfox.constants.marketplaces import RANDGALLERY_COM
from fetchfox.dtos import (
    AssetDTO,
    FloorDTO,
    HoldingDTO,
    ListingDTO,
    SaleDTO,
    TransactionAssetDTO,
    TransactionDTO,
    TransactionInputOutputDTO,
)
from fetchfox.helpers.otp import generate_otp
from . import utils
from .exceptions import (
    InvalidAlgorandAccountException,
    InvalidAlgorandAssetIdException,
    InvalidAlgorandCollectionIdException,
)

logger = logging.getLogger(__name__)


class Algorand(Blockchain):
    def __init__(self):
        super().__init__(
            name=ALGORAND,
            currency=ALGO,
            logo="https://s2.coinmarketcap.com/static/img/coins/64x64/4030.png",
        )

    # checks
    def check_account(self, account: str, exception: bool = True) -> bool:
        valid = utils.is_account(account)

        if not valid and exception:
            raise InvalidAlgorandAccountException(account)

        return valid

    def check_asset_id(self, asset_id: str, exception: bool = True) -> bool:
        valid = utils.is_asset_id(asset_id)

        if not valid and exception:
            raise InvalidAlgorandAssetIdException(asset_id)

        return valid

    def check_collection_id(self, collection_id: str, exception: bool = True) -> bool:
        valid = utils.is_address(collection_id)

        if not valid and exception:
            raise InvalidAlgorandCollectionIdException(collection_id)

        return valid

    # url builders
    def explorer_url(self, *, address: str = None, collection_id: str = None, asset_id: str = None, tx_hash: str = None) -> str:
        if address:
            return f"https://allo.info/account/{address.upper()}"

        if asset_id:
            return f"https://allo.info/asset/{asset_id}/nft"

        if collection_id:
            return f"https://allo.info/account/{collection_id.upper()}"

        if tx_hash:
            return f"https://allo.info/tx/{tx_hash.upper()}"

        return None

    def marketplace_url(self, *, collection_id: str = None, asset_id: str = None) -> str:
        if asset_id:
            return f"https://www.randgallery.com/algo-collection/?address={asset_id}"

        if collection_id:
            return f"https://randgallery.com/algo-collection/?address={collection_id}"

        return None

    # accounts
    @ttl_cache(ttl=600)
    def get_account_main_address(self, account: str, exception: bool = False) -> Optional[str]:
        if not self.check_account(account, exception=exception):
            return None

        resolved = self.resolve_account_name(account)

        return resolved or account

    def get_account_name(self, account: str) -> str:
        if utils.is_nf_domain(account):
            return account

        if utils.is_address(account):
            return nfdomains.get_nf_domain(account)

        return None

    def resolve_account_name(self, name: str) -> str:
        if utils.is_nf_domain(name):
            return nfdomains.resolve_nf_domain(name)

        return None

    # account attributes
    def get_account_assets(self, account: str, collection_id: str = None) -> Iterable[HoldingDTO]:
        address = self.get_account_main_address(account, exception=True)

        for holding in algonodecloud.get_account_assets(address):
            yield HoldingDTO(
                collection_id=None,
                asset_id=holding.asset_id,
                address=account,
                amount=holding.amount,
            )

    def get_account_balance(self, account: str) -> Tuple[float, str]:
        address = self.get_account_main_address(account, exception=True)

        balance = algonodecloud.get_account_balance(address)

        return balance, self.currency

    def get_account_transactions(self, account: str, last: int = 10) -> Iterable[TransactionDTO]:
        address = self.get_account_main_address(account, exception=True)

        for index, transaction in enumerate(algonodecloud.get_account_transactions(address)):
            if index >= last:
                break

            if transaction.tx_type == "pay":
                amount = transaction.payment_transaction.amount / 10**6
                receiver = transaction.payment_transaction.receiver
                unit = self.currency
            elif transaction.tx_type == "axfer":
                amount = transaction.asset_transfer_transaction.amount
                receiver = transaction.asset_transfer_transaction.receiver
                unit = transaction.asset_transfer_transaction.asset_id
            else:
                continue

            yield TransactionDTO(
                blockchain=self.name,
                address=address,
                tx_hash=transaction.id,
                message=transaction.note,
                inputs=[
                    TransactionInputOutputDTO(
                        address=transaction.sender,
                        assets=[
                            TransactionAssetDTO(
                                amount=amount,
                                unit=unit,
                            ),
                        ],
                    ),
                ],
                outputs=[
                    TransactionInputOutputDTO(
                        address=receiver,
                        assets=[
                            TransactionAssetDTO(
                                amount=amount,
                                unit=unit,
                            ),
                        ],
                    ),
                ],
            )

    # assets
    def get_asset(self, collection_id: str, asset_id: str, fetch_metadata: bool = True, *args, **kwargs) -> AssetDTO:
        if collection_id:
            self.check_collection_id(collection_id)

        self.check_asset_id(asset_id)

        asset_data = algonodecloud.get_asset_data(asset_id)

        if fetch_metadata:
            metadata = algonodecloud.get_asset_metadata(asset_id)
            metadata["name"] = asset_data.params.name
        else:
            metadata = {}

        return AssetDTO(
            collection_id=asset_data.params.creator,
            asset_id=asset_id,
            metadata=metadata,
        )

    def get_asset_holders(self, collection_id: str, asset_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        self.check_collection_id(collection_id)
        self.check_asset_id(asset_id)

        for holder in algonodecloud.get_asset_holders(asset_id):
            yield HoldingDTO(
                collection_id=collection_id,
                asset_id=holder.asset_id,
                address=holder.address,
                amount=holder.amount,
            )

    # collections
    def get_collection_assets(self, collection_id: str, fetch_metadata: bool = True, *args, **kwargs) -> Iterable[AssetDTO]:
        self.check_collection_id(collection_id)

        collection_assets = algonodecloud.get_account_created_assets(collection_id)

        for asset_id in collection_assets:
            if fetch_metadata:
                yield self.get_asset(
                    collection_id=collection_id,
                    asset_id=asset_id,
                )
            else:
                yield AssetDTO(
                    collection_id=collection_id,
                    asset_id=asset_id,
                    metadata={},
                )

    def get_collection_floor(self, collection_id: str, *args, **kwargs) -> FloorDTO:
        self.check_collection_id(collection_id)

        floor = None
        count = 0

        for listing in self.get_collection_listings(collection_id):
            count += 1

            if floor is None:
                floor = listing
            elif listing.usd < floor.usd:
                floor = listing

        return FloorDTO(
            listing=floor,
            listing_count=count,
        )

    def get_collection_listings(self, collection_id: str, *args, **kwargs) -> Iterable[ListingDTO]:
        self.check_collection_id(collection_id)

        for listing in randswapcom.get_creator_listings(collection_id):
            asset_id = str(listing.asset_id)
            asset_ids = [asset_id]
            asset_names = [""]

            listed_at = datetime.fromtimestamp(
                listing.timestamp // 1000,
            ).replace(
                tzinfo=pytz.UTC,
            )

            yield ListingDTO(
                identifier=listing.timestamp,
                collection_id=collection_id,
                asset_ids=asset_ids,
                asset_names=asset_names,
                listing_id=listing.timestamp,
                marketplace=RANDGALLERY_COM,
                price=listing.price,
                currency=ALGO,
                listed_at=listed_at,
                listed_by=listing.seller_address,
                marketplace_url=listing.url,
            )

    def get_collection_sales(self, collection_id: str, *args, **kwargs) -> Iterable[SaleDTO]:
        return []

    def get_collection_holders(self, collection_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        self.check_collection_id(collection_id)

        for asset in self.get_collection_assets(collection_id, fetch_metadata=False):
            yield from self.get_asset_holders(collection_id, asset.asset_id)

    def get_collection_supply(self, collection_id: str, *args, **kwargs) -> int:
        self.check_collection_id(collection_id)

        return algonodecloud.get_account_created_supply(
            creator_address=collection_id,
        )

    # others
    def get_otp(self) -> float:
        return float(f"0.{generate_otp()}")

    def verify_otp(self, account: str, otp: float) -> TransactionDTO:
        return super().verify_otp(account.upper(), otp)

    # sanity
    def sanity_check(self) -> bool:
        collection_id = "W2FPJKZZ6QP4PPAOJZWEUMMA372FF3MJD3PY4GNBZTTR3XKG464INIXMIQ"

        asset_id = "1180518311"
        asset_name, asset_number = "Foundations", 471

        self.validate_sanity_check(collection_id, asset_id, asset_name, asset_number)



from fetchfox.constants.algorand.basics import ALGORAND
from fetchfox.constants.cardano.basics import CARDANO
from fetchfox.constants.ethereum.basics import ETHEREUM
from fetchfox.constants.polygon.basics import POLYGON


BLOCKCHAINS = [
    ALGORAND,
    CARDANO,
    ETHEREUM,
    POLYGON,
]



# collections
from .collections import get_listings as get_collection_listings
from .collections import get_slug as get_collection_slug

# events
from .events import get_sales as get_collection_sales



# accounts
from .accounts import get_assets as get_account_assets
from .accounts import get_balance as get_account_balance
from .accounts import get_created_assets as get_account_created_assets
from .accounts import get_created_supply as get_account_created_supply
from .accounts import get_transactions as get_account_transactions

# assets
from .assets import get_data as get_asset_data
from .assets import get_holders as get_asset_holders
from .assets import get_metadata as get_asset_metadata



from typing import Iterable

from fetchfox.checks import check_str
from .common import get
from .schemas import (
    AccountAssetsSchema,
    AccountTransactionSchema,
)


def get_assets(address: str) -> Iterable[AccountAssetsSchema]:
    check_str(address, "algonodecloud.address")
    address = address.strip().upper()

    next_token = None

    while True:
        response, status_code = get(
            service=f"accounts/{address}/assets",
            params={
                "include-all": "false",
                "next": next_token,
            },
        )

        yield from filter(
            lambda a: a.amount > 0,
            map(
                AccountAssetsSchema.model_validate,
                response["assets"],
            ),
        )

        next_token = response.get("next-token")

        if not next_token:
            break


def get_created_assets(address: str) -> Iterable[int]:
    check_str(address, "algonodecloud.address")
    address = address.strip().upper()

    next_token = None

    while True:
        response, status_code = get(
            service=f"accounts/{address}/created-assets",
            params={
                "include-all": "false",
                "next": next_token,
            },
        )

        for asset in response["assets"]:
            yield str(asset["index"])

        next_token = response.get("next-token")

        if not next_token:
            break


def get_created_supply(creator_address: str) -> int:
    return len(list(get_created_assets(creator_address)))


def get_balance(address: str) -> float:
    check_str(address, "algonodecloud.address")
    address = address.strip().upper()

    response, status_code = get(
        service=f"accounts/{address}",
        params={
            "include-all": "false",
        },
    )

    return int(response["account"]["amount"]) / 10**6


def get_transactions(address: str) -> Iterable[AccountTransactionSchema]:
    check_str(address, "algonodecloud.address")
    address = address.strip().upper()

    next_token = None

    while True:
        response, status_code = get(
            service=f"accounts/{address}/transactions",
            params={
                "next": next_token,
            },
        )

        yield from map(
            AccountTransactionSchema.model_validate,
            response["transactions"],
        )

        next_token = response.get("next-token")

        if not next_token:
            break



import os
from typing import Tuple

from fetchfox import rest

BASE_URL_DEFAULT = "https://server.jpgstoreapis.com"
BASE_URL = os.getenv("JPGSTORE_BASE_URL") or BASE_URL_DEFAULT


def get(service: str, params: dict = None, headers: dict = None) -> Tuple[dict, int]:
    return rest.get(
        url=f"{BASE_URL}/{service}",
        params=params,
        headers=headers,
        sleep=5,
    )



import os
from typing import Tuple

from fetchfox import rest

BASE_URL_DEFAULT = "https://mainnet-idx.algonode.cloud"
BASE_URL = os.getenv("ALGONODECLOUD_API_BASE_URL") or BASE_URL_DEFAULT


def get(service: str, params: dict = None, version: int = 2) -> Tuple[dict, int]:
    return rest.get(
        url=f"{BASE_URL}/v{version}/{service}",
        params=params,
    )



from base64 import b64decode
from typing import Optional

from pydantic import Field, field_validator

from fetchfox.apis.base import BaseModel


class AccountAssetsSchema(BaseModel):
    amount: int
    asset_id: str = Field(..., alias="asset-id")

    @field_validator("asset_id", mode="before")
    def convert_asset_id(cls, value):
        return str(value)


class AssetsDataSchema(BaseModel):
    class AssetDataParamsSchema(BaseModel):
        creator: str
        name: str

    index: str
    params: AssetDataParamsSchema

    @field_validator("index", mode="before")
    def convert_index(cls, value):
        return str(value)


class AssetHolderSchema(BaseModel):
    address: str
    amount: int
    asset_id: str = Field(..., alias="asset-id")

    @field_validator("asset_id", mode="before")
    def convert_asset_id(cls, value):
        return str(value)


class AccountTransactionSchema(BaseModel):
    class AssetTransferTransactionSchema(BaseModel):
        amount: float
        asset_id: str = Field(..., alias="asset-id")
        receiver: str

        @field_validator("asset_id", mode="before")
        def convert_asset_id(cls, value):
            return str(value)

    class PaymentTransactionSchema(BaseModel):
        amount: float
        receiver: str

    id: str
    note: Optional[str] = None
    sender: str
    tx_type: str = Field(..., alias="tx-type")
    asset_transfer_transaction: Optional[AssetTransferTransactionSchema] = Field(None, alias="asset-transfer-transaction")
    payment_transaction: Optional[PaymentTransactionSchema] = Field(None, alias="payment-transaction")

    @field_validator("note", mode="before")
    def decode_note(cls, value):
        if value is None:
            return None

        return b64decode(value).decode("utf-8")



from ..exceptions import (
    InvalidCollectionIdException,
    InvalidAssetIdException,
    InvalidWalletException,
)


class InvalidEvmAccountException(InvalidWalletException):
    def __init__(self, string: str, blockchain: str = "evm"):
        super().__init__(f"'{string}' is not a valid {blockchain} address or ens domain.")


class InvalidEvmCollectionIdException(InvalidCollectionIdException):
    def __init__(self, string: str, blockchain: str = "evm"):
        super().__init__(f"'{string}' is not a valid {blockchain} contract address.")


class InvalidEvmAssetIdException(InvalidAssetIdException):
    def __init__(self, string: str, blockchain: str = "evm"):
        super().__init__(f"'{string}' is not a valid {blockchain} asset id.")



from .common import get
from .schemas import TransactionSchema


def get_transaction(tx_hash: str, api_key: str = None, preprod: bool = False) -> TransactionSchema:
    response, status_code = get(
        service=f"transactions/{tx_hash}",
        api_key=api_key,
        preprod=preprod,
        check="data",
    )

    return TransactionSchema.model_validate(response["data"])



import logging
from datetime import datetime
from typing import Iterable, List, Optional, Tuple

import pytz
from cachetools.func import ttl_cache

from fetchfox.apis.evm import moralisio, ensideascom, openseaio
from fetchfox.blockchains.base import Blockchain
from fetchfox.constants.marketplaces import OPENSEA_IO
from fetchfox.dtos import (
    AssetDTO,
    FloorDTO,
    HoldingDTO,
    ListingDTO,
    SaleDTO,
    TransactionDTO,
    TransactionInputOutputDTO,
    TransactionAssetDTO,
)
from . import utils
from .exceptions import (
    InvalidEvmAssetIdException,
    InvalidEvmCollectionIdException,
    InvalidEvmAccountException,
)

logger = logging.getLogger(__name__)


class Evm(Blockchain):
    def __init__(
        self,
        name: str,
        currency: str,
        logo: str,
        moralisio_api_key: str = None,
        openseaio_api_key: str = None,
        preprod: bool = False,
    ):
        super().__init__(name, currency, logo, preprod=preprod)

        self.moralisio_api_key: str = moralisio_api_key
        self.openseaio_api_key: str = openseaio_api_key

    # checks
    def check_account(self, account: str, exception: bool = True) -> bool:
        valid = utils.is_account(account)

        if not valid and exception:
            raise InvalidEvmAccountException(account, self.name)

        return valid

    def check_asset_id(self, asset_id: str, exception: bool = True) -> bool:
        valid = utils.is_asset_id(asset_id)

        if not valid and exception:
            raise InvalidEvmAssetIdException(asset_id, self.name)

        return valid

    def check_collection_id(self, collection_id: str, exception: bool = True) -> bool:
        valid = utils.is_address(collection_id)

        if not valid and exception:
            raise InvalidEvmCollectionIdException(collection_id, self.name)

        return valid

    # accounts
    @ttl_cache(ttl=600)
    def get_account_main_address(self, account: str, exception: bool = False) -> Optional[str]:
        if not self.check_account(account, exception=exception):
            return None

        resolved = self.resolve_account_name(account)

        return resolved or account

    def get_account_name(self, account: str) -> str:
        if utils.is_ens_domain(account):
            return account

        if utils.is_address(account):
            return ensideascom.get_ens_domain(account)

        return None

    def resolve_account_name(self, name: str) -> str:
        if utils.is_ens_domain(name):
            return ensideascom.resolve_ens_domain(name)

        return None

    # account attributes
    def get_account_assets(self, account: str, collection_id: str = None) -> Iterable[HoldingDTO]:
        address = self.get_account_main_address(account, exception=True)

        account_assets = moralisio.get_address_assets(
            address,
            blockchain=self.name,
            api_key=self.moralisio_api_key,
            preprod=self.preprod,
        )

        for asset in account_assets:
            yield HoldingDTO(
                collection_id=asset.token_address,
                asset_id=asset.token_id,
                address=address,
                amount=asset.amount,
            )

    def get_account_balance(self, account: str) -> Tuple[float, str]:
        address = self.get_account_main_address(account, exception=True)

        balance = moralisio.get_address_balance(
            address,
            blockchain=self.name,
            api_key=self.moralisio_api_key,
            preprod=self.preprod,
        )

        return balance, self.currency

    def get_account_transactions(self, account: str, last: int = 10) -> Iterable[TransactionDTO]:
        def parse_transfer(ttype: str, tlist: list):
            for transfer in tlist:
                if ttype == "erc20":
                    amount = transfer.value
                    unit = transfer.token_symbol
                elif ttype == "nft":
                    amount = transfer.amount
                    unit = transfer.token_address + "/" + transfer.token_id
                else:
                    decimals = 18
                    amount = transfer.value / 10**decimals
                    unit = transfer.token_symbol

                tx_input = TransactionInputOutputDTO(
                    address=transfer.from_address,
                    assets=[
                        TransactionAssetDTO(
                            amount=amount,
                            unit=unit,
                        ),
                    ],
                )

                tx_output = TransactionInputOutputDTO(
                    address=transfer.to_address,
                    assets=[
                        TransactionAssetDTO(
                            amount=amount,
                            unit=unit,
                        ),
                    ],
                )

                yield tx_input, tx_output

        address = self.get_account_main_address(account, exception=True)

        account_transactions = moralisio.get_wallet_transactions(
            address,
            blockchain=self.name,
            api_key=self.moralisio_api_key,
            preprod=self.preprod,
        )

        for index, transaction in enumerate(account_transactions):
            if index >= last:
                break

            inputs, outputs = [], []

            transfers = {
                "erc20": transaction.erc20_transfers,
                "nft": transaction.nft_transfers,
                "native": transaction.native_transfers,
            }

            for transfer_type, transfer_list in transfers.items():
                for tx_input, tx_output in parse_transfer(transfer_type, transfer_list):
                    inputs.append(tx_input)
                    outputs.append(tx_output)

            yield TransactionDTO(
                blockchain=self.name,
                address=address,
                tx_hash=transaction.hash,
                message=transaction.summary,
                inputs=inputs,
                outputs=outputs,
            )

    # assets
    def get_asset(self, collection_id: str, asset_id: str, *args, **kwargs) -> AssetDTO:
        self.check_collection_id(collection_id)
        self.check_asset_id(asset_id)

        asset_data = moralisio.get_asset_data(
            contract_address=collection_id,
            asset_id=asset_id,
            blockchain=self.name,
            api_key=self.moralisio_api_key,
            preprod=self.preprod,
        )

        return AssetDTO(
            collection_id=collection_id,
            asset_id=asset_id,
            metadata=asset_data["metadata"],
        )

    def get_asset_holders(self, collection_id: str, asset_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        self.check_collection_id(collection_id)
        self.check_asset_id(asset_id)

        asset_holders = moralisio.get_asset_holders(
            collection_id,
            asset_id=asset_id,
            blockchain=self.name,
            api_key=self.moralisio_api_key,
            preprod=self.preprod,
        )

        for item in asset_holders:
            yield HoldingDTO(
                collection_id=item.token_address,
                asset_id=item.token_id,
                address=item.owner_of,
                amount=item.amount,
            )

    # collections
    def get_collection_assets(self, collection_id: str, fetch_metadata: bool = True, *args, **kwargs) -> Iterable[AssetDTO]:
        self.check_collection_id(collection_id)

        if fetch_metadata:
            asset_id = -1

            while True:
                try:
                    asset_id += 1

                    yield self.get_asset(
                        collection_id=collection_id,
                        asset_id=str(asset_id),
                    )
                except ValueError:
                    raise
                except Exception as exc:
                    break
        else:
            collection_assets = moralisio.get_contract_assets(
                contract_address=collection_id,
                blockchain=self.name,
                api_key=self.moralisio_api_key,
                preprod=self.preprod,
            )

            for asset_id in collection_assets:
                yield AssetDTO(
                    collection_id=collection_id,
                    asset_id=asset_id,
                    metadata={},
                )

    def get_collection_floor(self, collection_id: str, *args, **kwargs) -> FloorDTO:
        self.check_collection_id(collection_id)

        floor = None
        count = 0

        for listing in self.get_collection_listings(collection_id):
            count += 1

            if floor is None:
                floor = listing
            elif listing.usd < floor.usd:
                floor = listing

        return FloorDTO(
            listing=floor,
            listing_count=count,
        )

    def get_collection_listings(self, collection_id: str, slug: str = None, *args, **kwargs) -> Iterable[ListingDTO]:
        self.check_collection_id(collection_id)

        collection_listings = openseaio.get_collection_listings(
            collection_id,
            blockchain=self.name,
            api_key=self.openseaio_api_key,
            slug=slug,
        )

        for listing in collection_listings:
            asset_ids = []
            asset_names = []

            for offer in listing.protocol_data.parameters.offer:
                if offer.token.lower() != collection_id.lower():
                    continue

                asset_ids.append(offer.identifier_or_criteria)
                asset_names.append("")

            if listing.protocol_data.parameters.start_time:
                listed_at = datetime.utcfromtimestamp(int(listing.protocol_data.parameters.start_time))
            else:
                listed_at = datetime.now(tz=pytz.utc)

            marketplace_url = self.marketplace_url(
                collection_id=collection_id,
                asset_id=asset_ids[0],
            )

            yield ListingDTO(
                identifier=listing.order_hash,
                collection_id=collection_id,
                asset_ids=asset_ids,
                asset_names=asset_names,
                listing_id=listing.order_hash,
                marketplace=OPENSEA_IO,
                price=float(listing.price.current.value / 10**listing.price.current.decimals),
                currency=listing.price.current.currency.replace("WETH", "ETH"),
                listed_at=listed_at,
                listed_by=listing.protocol_data.parameters.offerer,
                tx_hash=listing.order_hash,
                marketplace_url=marketplace_url,
            )

    def get_collection_sales(self, collection_id: str, slug: str = None, *args, **kwargs) -> Iterable[SaleDTO]:
        self.check_collection_id(collection_id)

        collection_sale = openseaio.get_collection_sales(
            collection_id,
            blockchain=self.name,
            api_key=self.openseaio_api_key,
            slug=slug,
        )

        for sale in collection_sale:
            tx_hash = sale.transaction

            asset_id = sale.nft.identifier
            asset_name = sale.nft.name

            if sale.closing_date:
                confirmed_at = datetime.fromtimestamp(
                    sale.closing_date,
                    tz=pytz.utc,
                )
            else:
                confirmed_at = datetime.now(
                    tz=pytz.utc,
                )

            marketplace_url = self.marketplace_url(
                collection_id=collection_id,
                asset_id=asset_id,
            )

            explorer_url = self.explorer_url(
                tx_hash=tx_hash,
            )

            yield SaleDTO(
                identifier=f"{tx_hash}/{asset_id}",
                collection_id=collection_id,
                asset_ids=[asset_id],
                asset_names=[asset_name],
                tx_hash=tx_hash,
                marketplace=OPENSEA_IO,
                price=sale.payment.quantity / 10**sale.payment.decimals,
                currency=sale.payment.symbol.replace("WETH", "ETH"),
                confirmed_at=confirmed_at,
                sold_by=sale.seller,
                bought_by=sale.buyer,
                sale_id=tx_hash,
                marketplace_url=marketplace_url,
                explorer_url=explorer_url,
            )

    def get_collection_supply(self, collection_id: str, *args, **kwargs) -> int:
        self.check_collection_id(collection_id)

        return moralisio.get_contract_supply(
            collection_id,
            blockchain=self.name,
            api_key=self.moralisio_api_key,
            preprod=self.preprod,
        )

    def get_collection_holders(self, collection_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        self.check_collection_id(collection_id)

        collection_owners = moralisio.get_contract_holders(
            collection_id,
            blockchain=self.name,
            api_key=self.moralisio_api_key,
            preprod=self.preprod,
        )

        for asset in collection_owners:
            yield HoldingDTO(
                collection_id=asset.token_address,
                asset_id=asset.token_id,
                address=asset.owner_of,
                amount=asset.amount,
            )

    # others
    def verify_otp(self, account: str, otp: float) -> TransactionDTO:
        return super().verify_otp(account.lower(), otp)

    # tokens
    def get_token_exchange(self, collection_id: str, asset_id: str) -> Optional[float]:
        self.check_collection_id(collection_id)
        self.check_asset_id(asset_id)

        return moralisio.get_token_price(
            collection_id,
            blockchain=self.name,
            api_key=self.moralisio_api_key,
            preprod=self.preprod,
        )

    # sanity
    def api_keys(self) -> List[str]:
        return [
            "m:{moralisio}".format(
                moralisio=self.moralisio_api_key[-3:] if self.moralisio_api_key else "none",
            ),
            "os:{openseaio}".format(
                openseaio=self.openseaio_api_key[-3:] if self.openseaio_api_key else "none",
            ),
        ]



from fetchfox.constants.cardano.tokens.beard import (
    BEARD_TOKEN_ASSET_ID,
    BEARD_TOKEN_ASSET_NAME,
    BEARD_TOKEN_POLICY_ID,
    BEARD_TOKEN_FINGERPRINT,
    BEARD_TOKEN_SYMBOL,
)

from .base import CardanoToken


class BeardToken(CardanoToken):
    def __init__(self, dexhunterio_partner_code: str = None):
        super().__init__(
            asset_id=BEARD_TOKEN_ASSET_ID,
            asset_name=BEARD_TOKEN_ASSET_NAME,
            fingerprint=BEARD_TOKEN_FINGERPRINT,
            policy_id=BEARD_TOKEN_POLICY_ID,
            symbol=BEARD_TOKEN_SYMBOL,
            decimals=0,
            taptools_pair_id="52bd1e5a642e2141e8fb7ea6fa996960bc91db0b040c6d35013cf3efecac2dfc",
            dexhunterio_partner_code=dexhunterio_partner_code,
        )



from .base import CardanoToken


class HoskyToken(CardanoToken):
    def __init__(self, dexhunterio_partner_code: str = None):
        super().__init__(
            asset_id="a0028f350aaabe0545fdcb56b039bfb08e4bb4d8c4d7c3c7d481c235484f534b59",
            asset_name="484f534b59",
            fingerprint="asset17q7r59zlc3dgw0venc80pdv566q6yguw03f0d9",
            policy_id="a0028f350aaabe0545fdcb56b039bfb08e4bb4d8c4d7c3c7d481c235",
            symbol="HOSKY",
            decimals=0,
            coingecko_id="hosky",
            taptools_pair_id="a449f98fa06fff1d6c17b95c65ca609660049dc683cc3e94d3846b05419fc20d",
            dexhunterio_partner_code=dexhunterio_partner_code,
        )

    @property
    def ada(self) -> float:
        return 0.0

    @property
    def usd(self) -> float:
        return 0.0



import logging
import warnings
from datetime import datetime
from typing import Iterable, List, Optional, Tuple

import pytz
from cachetools.func import ttl_cache

from fetchfox.apis.cardano import dexhunterio
from fetchfox.apis.cardano import gomaestroorg, jpgstore
from fetchfox.blockchains.base import Blockchain
from fetchfox.constants.blockchains import CARDANO
from fetchfox.constants.cardano.ranks import CNFT_RANKS
from fetchfox.constants.currencies import ADA
from fetchfox.constants.marketplaces import JPG_STORE
from fetchfox.dtos import (
    AssetDTO,
    FloorDTO,
    HoldingDTO,
    ListingDTO,
    RankDTO,
    SaleDTO,
    SaleType,
    TransactionAssetDTO,
    TransactionDTO,
    TransactionInputOutputDTO,
)
from fetchfox.helpers.otp import generate_otp
from . import utils
from .exceptions import (
    InvalidCardanoAccountException,
    InvalidCardanoAssetIdException,
    InvalidCardanoCollectionIdException,
)
from ..exceptions import UnavailableMetadataException

WINTER_NFT_ADDRESS = "addr1qxnrv2quqxhvwxtxmygsmkufph4kjju6j5len7k2ljslpz8ql7k7gehlfvj6ektgu9ns8yx8epcp66337khxeq82rpgqe6lqyk"
SHELLY_EPOCH = 1596491091
SHELLY_SLOT = 4924800

logger = logging.getLogger(__name__)


class Cardano(Blockchain):
    def __init__(
        self,
        preprod: bool = False,
        gomaestroorg_api_key: str = None,
        dexhunterio_partner_code: str = None,
    ):
        super().__init__(
            name=CARDANO,
            currency=ADA,
            logo="https://s2.coinmarketcap.com/static/img/coins/64x64/2010.png",
            preprod=preprod,
        )

        self.gomaestroorg_api_key: str = gomaestroorg_api_key
        self.dexhunterio_partner_code: str = dexhunterio_partner_code

    # checks
    def check_account(self, account: str, exception: bool = True) -> bool:
        if self.preprod:
            valid = utils.is_preprod_account(account)
        else:
            valid = utils.is_account(account)

        if not valid and exception:
            raise InvalidCardanoAccountException(account)

        return valid

    def check_asset_id(self, asset_id: str, exception: bool = True) -> bool:
        valid = utils.is_asset_id(asset_id)

        if not valid and exception:
            raise InvalidCardanoAssetIdException(asset_id)

        return valid

    def check_collection_id(self, collection_id: str, exception: bool = True) -> bool:
        valid = utils.is_policy_id(collection_id)

        if not valid and exception:
            raise InvalidCardanoCollectionIdException(collection_id)

        return valid

    # url builders
    def explorer_url(self, *, address: str = None, collection_id: str = None, asset_id: str = None, tx_hash: str = None) -> str:
        if address:
            return f"https://pool.pm/{address.lower()}"

        if asset_id:
            return f"https://cardanoscan.io/token/{asset_id.lower()}"

        if collection_id:
            return f"https://pool.pm/policy/{collection_id.lower()}"

        if tx_hash:
            return f"https://cardanoscan.io/transaction/{tx_hash.lower()}"

        return None

    def marketplace_url(self, collection_id: str = None, asset_id: str = None) -> str:
        if asset_id:
            return f"https://www.jpg.store/asset/{asset_id.lower()}"

        if collection_id:
            return f"https://jpg.store/collection/{collection_id.lower()}"

        return None

    # accounts
    def get_account_stake_address(self, account: str) -> str:
        warnings.deprecated("cardano.get_account_stake_address is deprecated, please use cardano.get_account_main_address instead")

        return self.get_account_main_address(account)

    @ttl_cache(ttl=600)
    def get_account_main_address(self, account: str, exception: bool = False) -> Optional[str]:
        if not self.check_account(account, exception=exception):
            return None

        is_stake_address = False

        if self.preprod:
            if utils.is_preprod_stake_address(account):
                is_stake_address = True
        else:
            if utils.is_stake_address(account):
                is_stake_address = True

        if is_stake_address:
            return account

        is_address = False

        if self.preprod:
            if utils.is_preprod_address(account):
                is_address = True
        else:
            if utils.is_address(account):
                is_address = True

        if is_address:
            return gomaestroorg.get_address_stake_address(
                account,
                api_key=self.gomaestroorg_api_key,
                preprod=self.preprod,
            )

        if utils.is_ada_handle(account):
            resolution = gomaestroorg.resolve_handle(
                account,
                api_key=self.gomaestroorg_api_key,
                preprod=self.preprod,
            )

            return resolution

        return None

    def get_account_name(self, account: str) -> str:
        if utils.is_ada_handle(account):
            return account

        stake_address = self.get_account_main_address(account)

        return gomaestroorg.get_handle(
            stake_address,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

    def resolve_account_name(self, name: str) -> str:
        if not utils.is_ada_handle(name):
            return None

        resolution = gomaestroorg.resolve_handle(
            name,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        if not resolution:
            return None

        return resolution["stake_address"]

    def format_account(self, account: str) -> str:
        self.check_account(account, exception=True)

        handle = self.get_account_name(account)

        if handle:
            punycode = handle.encode().decode("idna")

            if handle != punycode:
                return f"{punycode} (${handle})"

            return f"${handle}"

        return super().format_account(account)

    # account attributes
    def get_account_assets(self, account: str, collection_id: str = None) -> Iterable[HoldingDTO]:
        stake_address = self.get_account_main_address(account, exception=True)

        if collection_id is not None:
            self.check_collection_id(collection_id)

        account_assets = gomaestroorg.get_account_assets(
            stake_address,
            policy_id=collection_id,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        for account_asset in account_assets:
            asset_id = account_asset.unit
            amount = account_asset.amount

            policy_id, _, _ = utils.split_asset_id(asset_id)

            yield HoldingDTO(
                collection_id=policy_id,
                asset_id=asset_id,
                address=stake_address,
                amount=amount,
            )

    def get_account_balance(self, account: str) -> Tuple[float, str]:
        stake_address = self.get_account_main_address(account, exception=True)

        balance = gomaestroorg.get_account_balance(
            stake_address,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        return balance, self.currency

    def get_account_transactions(self, account: str, last: int = 10) -> Iterable[TransactionDTO]:
        def parse_input_output(items: list) -> Iterable[TransactionInputOutputDTO]:
            for item in items:
                item_address = item.address
                item_assets = []

                for asset in item.assets:
                    unit = asset.unit
                    amount = asset.amount

                    if unit == "lovelace":
                        unit = self.currency
                        amount = amount / 10**6

                    item_assets.append(
                        TransactionAssetDTO(
                            amount=amount,
                            unit=unit,
                        )
                    )

                yield TransactionInputOutputDTO(
                    address=item_address,
                    assets=item_assets,
                )

        stake_address = self.get_account_main_address(account, exception=True)

        txs = []

        transactions = gomaestroorg.get_account_transactions(
            stake_address,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        for address, tx in transactions:
            txs.append((address, tx))

        sorted_transactions = list(
            sorted(
                txs,
                key=lambda x: x[1].block_timestamp,
                reverse=True,
            )
        )

        for address, tx in sorted_transactions[:last]:
            message = None

            if tx.metadata:
                if tx.metadata.m_674:
                    if tx.metadata.m_674.msg:
                        message = "\n".join(tx.metadata.m_674.msg)

            inputs = list(parse_input_output(tx.inputs))
            outputs = list(parse_input_output(tx.outputs))

            yield TransactionDTO(
                blockchain=self.name,
                address=stake_address,
                tx_hash=tx.tx_hash,
                message=message,
                inputs=inputs,
                outputs=outputs,
            )

    # assets
    def get_asset(self, collection_id: str, asset_id: str, *args, **kwargs) -> AssetDTO:
        if collection_id is None:
            collection_id = asset_id[:56]

        self.check_collection_id(collection_id)
        self.check_asset_id(asset_id)

        asset_data = gomaestroorg.get_asset_data(
            asset_id,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        if not asset_data:
            raise UnavailableMetadataException(asset_id)

        return AssetDTO(
            collection_id=collection_id,
            asset_id=asset_id,
            metadata=asset_data,
            supply=int(asset_data.get("total_supply", 1)),
        )

    def get_asset_holders(self, collection_id: str, asset_id: str, *args, **kwargs) -> Iterable[HoldingDTO]:
        self.check_collection_id(collection_id)
        self.check_asset_id(asset_id)

        asset_owners = gomaestroorg.get_asset_holders(
            asset_id,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        for owner in asset_owners:
            yield HoldingDTO(
                collection_id=collection_id,
                asset_id=owner.asset_id,
                address=owner.account,
                amount=owner.amount,
            )

    def get_asset_rank(self, collection_id: str, asset_id: str, *args, **kwargs) -> RankDTO:
        self.check_collection_id(collection_id)
        self.check_asset_id(asset_id)

        _, _, decoded_asset_name = utils.split_asset_id(asset_id)

        if collection_id not in CNFT_RANKS:
            return None

        if not decoded_asset_name:
            return None

        if decoded_asset_name not in CNFT_RANKS[collection_id]:
            return None

        digits = "".join((c for c in decoded_asset_name if c.isdigit()))

        if not digits:
            return None

        number = int(digits)

        return RankDTO(
            collection_id=collection_id,
            number=number,
            asset_id=asset_id,
            rank=CNFT_RANKS[collection_id][decoded_asset_name],
        )

    # collections
    def get_collection_assets(
        self,
        collection_id: str,
        discriminator: str = None,
        fetch_metadata: bool = True,
        *args,
        **kwargs,
    ) -> Iterable[AssetDTO]:
        self.check_collection_id(collection_id)

        if discriminator:
            discriminator = discriminator.lower()

        collection_assets = gomaestroorg.get_policy_assets(
            collection_id,
            discriminator=discriminator,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        for asset in collection_assets:
            asset_name = asset.asset_name.lower()

            if not asset_name:  # skip royalty token
                continue

            asset_id = f"{collection_id}{asset_name}"
            cip25_metadata = asset.asset_standards.cip25_metadata
            cip68_metadata = asset.asset_standards.cip68_metadata

            metadata = cip68_metadata or cip25_metadata

            if not metadata:
                continue

            yield AssetDTO(
                collection_id=collection_id,
                asset_id=asset_id,
                metadata=metadata,
                supply=asset.total_supply,
            )

    def get_collection_lock_date(self, collection_id: str, *args, **kwargs) -> datetime:
        slot = gomaestroorg.get_policy_lock_slot(
            collection_id,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        if slot is None:
            return None

        epoch = SHELLY_EPOCH + (slot - SHELLY_SLOT)
        return datetime.utcfromtimestamp(epoch).replace(tzinfo=pytz.UTC)

    def get_collection_floor(self, collection_id: str, discriminator: str = None, *args, **kwargs) -> FloorDTO:
        self.check_collection_id(collection_id)

        collection_listings = self.get_collection_listings(
            collection_id,
            discriminator=discriminator,
        )

        floor = None
        count = 0

        for listing in collection_listings:
            count += 1

            if floor is None:
                floor = listing
            elif listing.usd < floor.usd:
                floor = listing

        return FloorDTO(
            listing=floor,
            listing_count=count,
        )

    def get_collection_listings(self, collection_id: str, discriminator: str = None, *args, **kwargs) -> Iterable[ListingDTO]:
        self.check_collection_id(collection_id)

        if discriminator:
            discriminator = discriminator.lower()

        collection_listings = jpgstore.get_policy_listings(collection_id)

        for listing in collection_listings:
            asset_id = listing.asset_id
            policy_id, asset_name, _ = utils.split_asset_id(asset_id)

            # required for multi-book policies (e.g. monsters, greek classics)
            if discriminator:
                if not asset_name.startswith(discriminator):
                    continue

            asset_ids = []
            asset_names = []

            if listing.listing_type == "BUNDLE":
                for bundled_asset in listing.bundled_assets:
                    asset_ids.append(bundled_asset.asset_id)
                    asset_names.append(bundled_asset.display_name)
            else:
                asset_ids.append(listing.asset_id)
                asset_names.append(listing.display_name)

            if listing.confirmed_at:
                listed_at = datetime.fromisoformat(listing.confirmed_at.replace("Z", "+00:00"))
            else:
                listed_at = datetime.now(tz=pytz.utc)

            yield ListingDTO(
                identifier=listing.tx_hash,
                collection_id=policy_id,
                asset_ids=asset_ids,
                asset_names=asset_names,
                listing_id=listing.listing_id,
                marketplace=JPG_STORE,
                price=listing.price_lovelace // 10**6,
                currency=self.currency,
                listed_at=listed_at,
                listed_by=None,
                tx_hash=listing.tx_hash,
                marketplace_url=listing.url,
            )

    def get_collection_ranks(self, collection_id: str, *args, **kwargs) -> Iterable[RankDTO]:
        self.check_collection_id(collection_id)

        if collection_id not in CNFT_RANKS:
            return

        for asset_name, rank in CNFT_RANKS[collection_id].items():
            number = int("".join((c for c in asset_name if c.isdigit())))

            yield RankDTO(
                collection_id=collection_id,
                number=number,
                asset_id=None,
                rank=int(rank),
            )

    def get_collection_sales(self, collection_id: str, discriminator: str = None, *args, **kwargs) -> Iterable[SaleDTO]:
        self.check_collection_id(collection_id)

        if discriminator:
            discriminator = discriminator.lower()

        collection_sales = jpgstore.get_collection_sales(collection_id)

        for sale in collection_sales:
            tx_hash = sale.tx_hash

            asset_id = sale.asset_id
            policy_id, asset_name, encoded_asset_name = utils.split_asset_id(asset_id)

            # required for multi-book policies (e.g. monsters, greek classics)
            if discriminator:
                if not asset_name.startswith(discriminator):
                    continue

            bulk_size = sale.bulk_size

            if sale.action == "ACCEPT_OFFER":
                buyer = sale.seller_address
                seller = sale.signer_address
                sale_type = SaleType.OFFER
            elif sale.action == "ACCEPT_COLLECTION_OFFER":
                buyer = sale.signer_address
                seller = sale.seller_address
                sale_type = SaleType.COLLECTION_OFFER
            elif sale.action == "BUY":
                buyer = sale.signer_address
                seller = sale.seller_address
                sale_type = SaleType.PURCHASE
            else:
                continue

            asset_ids = []
            asset_names = []

            if sale.listing_from_tx_history.bundled_assets:
                for bundled_asset in sale.listing_from_tx_history.bundled_assets:
                    asset_ids.append(bundled_asset.asset_id)
                    asset_names.append(bundled_asset.display_name)
            else:
                asset_ids.append(sale.asset_id)
                asset_names.append(sale.display_name)

            if sale.confirmed_at:
                confirmed_at = datetime.fromisoformat(sale.confirmed_at.replace("Z", "+00:00"))
            else:
                confirmed_at = datetime.now(tz=pytz.utc)

            yield SaleDTO(
                identifier=tx_hash,
                collection_id=policy_id,
                asset_ids=asset_ids,
                asset_names=asset_names,
                tx_hash=tx_hash,
                marketplace=JPG_STORE,
                price=sale.amount_lovelace // 10**6,
                currency=self.currency,
                confirmed_at=confirmed_at,
                type=sale_type,
                bulk_size=bulk_size,
                sold_by=seller,
                bought_by=buyer,
                marketplace_url=sale.url,
                explorer_url=f"https://cardanoscan.io/transaction/{tx_hash}",
            )

    def get_collection_supply(self, collection_id: str, discriminator: str = None, *args, **kwargs) -> int:
        self.check_collection_id(collection_id)

        if discriminator:
            return gomaestroorg.get_policy_asset_count(
                collection_id,
                discriminator=discriminator,
                api_key=self.gomaestroorg_api_key,
                preprod=self.preprod,
            )

        return gomaestroorg.get_policy_supply(
            collection_id,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

    def get_collection_holders(self, collection_id: str, discriminator: str = None, max_threads: int = 3, *args, **kwargs) -> Iterable[HoldingDTO]:
        self.check_collection_id(collection_id)

        collection_owners = gomaestroorg.get_policy_holders(
            collection_id,
            api_key=self.gomaestroorg_api_key,
            preprod=self.preprod,
        )

        for owner in collection_owners:
            stake_address = owner.account

            for asset in owner.assets:
                encoded_asset_name = asset.name.lower()
                asset_id = f"{collection_id}{encoded_asset_name}"
                amount = asset.amount

                # required for multi-book policies (e.g. monsters, greek classics)
                if discriminator:
                    if not encoded_asset_name.startswith(discriminator.lower()):
                        continue

                yield HoldingDTO(
                    collection_id=collection_id,
                    asset_id=asset_id,
                    address=stake_address,
                    amount=amount,
                )

    # others
    def get_otp(self) -> float:
        return float(f"1.{generate_otp()}")

    def verify_otp(self, account: str, otp: float) -> TransactionDTO:
        return super().verify_otp(account.lower(), otp)

    # tokens
    def get_token_exchange(self, collection_id: str, asset_id: str) -> Optional[float]:
        self.check_asset_id(asset_id)

        if collection_id is not None:
            self.check_collection_id(collection_id)

        return dexhunterio.get_asset_average_price(
            asset_id,
            partner_code=self.dexhunterio_partner_code,
        )

    # sanity
    def sanity_check(self) -> bool:
        if self.preprod:
            collection_id = "cd91ae78fc25a1af2091893e59363189cc426d02aaaaf5312d810fc6"
            asset_id = "cd91ae78fc25a1af2091893e59363189cc426d02aaaaf5312d810fc6416c696365496e576f6e6465726c616e64417564696f626f6f6b303335"
            asset_name, asset_number = "Alice in Wonderland", 35
        else:
            collection_id = "9cb921b32bfe214a739ed824f3f2da4e16c535a5448253d2951cc732"
            asset_id = "9cb921b32bfe214a739ed824f3f2da4e16c535a5448253d2951cc732416c696365496e576f6e6465726c616e64303335"
            asset_name, asset_number = "Alice's Adventures in Wonderland", 35

        self.validate_sanity_check(collection_id, asset_id, asset_name, asset_number)

    def api_keys(self) -> List[str]:
        return [
            "m:{gomaestroorg}".format(
                gomaestroorg=self.gomaestroorg_api_key[-3:] if self.gomaestroorg_api_key else "none",
            ),
            "dh:{dexhunterio}".format(
                dexhunterio=self.dexhunterio_partner_code[-3:] if self.dexhunterio_partner_code else "none",
            ),
        ]



from typing import List, Tuple

from cachetools.func import ttl_cache

from fetchfox.apis import coingeckocom
from fetchfox.apis import price
from fetchfox.apis.cardano import dexhunterio
from fetchfox.constants.currencies import ADA


class CardanoToken:
    def __init__(
        self,
        asset_id: str,
        asset_name: str,
        fingerprint: str,
        policy_id: str,
        symbol: str,
        decimals: int = 6,
        coingecko_id: str = None,
        taptools_pair_id: str = None,
        dexhunterio_partner_code: str = None,
    ):
        self.asset_id: str = asset_id
        self.asset_name: str = asset_name
        self.fingerprint: str = fingerprint
        self.policy_id: str = policy_id
        self.symbol: str = symbol
        self.decimals: int = decimals

        self.coingecko_id: str = coingecko_id
        self.taptools_pair_id: str = taptools_pair_id
        self.dexhunterio_partner_code = dexhunterio_partner_code

    @property
    @ttl_cache(ttl=30)
    def ada(self) -> float:
        return dexhunterio.get_asset_average_price(
            self.asset_id,
            partner_code=self.dexhunterio_partner_code,
        )

    @property
    @ttl_cache(ttl=30)
    def usd(self) -> float:
        return price.usd(ADA) * self.ada

    @property
    @ttl_cache(ttl=30)
    def ath_usd(self) -> float:
        return price.ath_usd(self.symbol)

    def history(self, days: int = 7) -> List[Tuple[int, float]]:
        for exchange in coingeckocom.get_exchange_history(self.symbol, days=days):
            yield exchange.timestamp, exchange.price

    @property
    def cardanoscan_url(self):
        return f"https://cardanoscan.io/token/{self.asset_id}"

    @property
    def cexplorer_url(self):
        return f"https://cexplorer.io/asset/{self.fingerprint}"

    @property
    def coingecko_url(self):
        return f"https://www.coingecko.com/en/coins/{self.coingecko_id}"

    @property
    def dexhunter_url(self):
        return f"https://app.dexhunter.io/trends?asset={self.asset_id}&status=COMPLETED"

    @property
    def minswap_url(self):
        return f"https://app.minswap.org/swap?currencySymbolA={self.policy_id}&tokenNameA={self.asset_name}&currencySymbolB=&tokenNameB="

    @property
    def taptools_url(self):
        return f"https://www.taptools.io/charts/token/f5808c2c990d86da54bfc97d89cee6efa20cd8461616359478d96b4c.{self.taptools_pair_id}"



from typing import Optional, List

from pydantic import field_validator

from fetchfox.apis.base import BaseModel


class AddressesAssetsSchema(BaseModel):
    amount: float
    name: str
    token_address: str
    token_id: str

    @field_validator("token_id", mode="before")
    def convert_token_id(cls, value):
        return str(value)


class AssetHoldersSchema(BaseModel):
    amount: float
    owner_of: str
    token_address: str
    token_id: str

    @field_validator("token_id", mode="before")
    def convert_token_id(cls, value):
        return str(value)


class NftHoldersSchema(BaseModel):
    amount: float
    owner_of: str
    token_address: str
    token_id: str

    @field_validator("token_id", mode="before")
    def convert_token_id(cls, value):
        return str(value)


class WalletTransactionsSchema(BaseModel):
    class Erc20TransfersSchema(BaseModel):
        from_address: str
        to_address: str
        token_decimals: int
        token_symbol: str
        value: float

    class NftTransfersSchema(BaseModel):
        from_address: str
        to_address: str
        amount: float
        token_address: str
        token_id: str

        @field_validator("token_id", mode="before")
        def convert_token_id(cls, value):
            return str(value)

    class NativeTransfersSchema(BaseModel):
        from_address: str
        to_address: str
        decimals: int = 18
        token_symbol: str
        value: float

    hash: str
    summary: Optional[str] = None
    erc20_transfers: List[Erc20TransfersSchema] = None
    nft_transfers: List[NftTransfersSchema] = None
    native_transfers: List[NativeTransfersSchema] = None



from fetchfox.constants.algorand.marketplaces import ALGOXNFT_COM, RANDGALLERY_COM, SHUFL_APP
from fetchfox.constants.cardano.marketplaces import JPG_STORE
from fetchfox.constants.evm.marketplaces import OPENSEA_IO, RARIBLE_COM


UNKNOWN = "unknown"


MARKETPLACES = [
    ALGOXNFT_COM,
    RANDGALLERY_COM,
    SHUFL_APP,
    JPG_STORE,
    OPENSEA_IO,
    RARIBLE_COM,
]



from pydantic import Field, field_validator

from fetchfox.apis.base import BaseModel


class CreatorListingSchema(BaseModel):
    asset_id: str = Field(..., alias="assetId")
    price: float
    seller_address: str = Field(..., alias="sellerAddress")
    timestamp: int

    @field_validator("asset_id", mode="before")
    def convert_asset_id(cls, value):
        return str(value)

    @property
    def url(self) -> str:
        return f"https://randgallery.com/algo-collection/?address={self.asset_id}"



import asyncio
import json
import logging
import ssl
from typing import Tuple

import backoff
import certifi
from aiohttp import TCPConnector, ClientSession, ClientTimeout
from aiohttp.client_exceptions import ContentTypeError

logger = logging.getLogger(__name__)


@backoff.on_exception(backoff.expo, exception=Exception, max_time=30)
def get(url: str, headers: dict = None, params: dict = None, sleep: float = 0.05, timeout: float = None, check: str = None) -> Tuple[dict, int]:
    headers = {name: value for name, value in (headers or {}).items() if value is not None}
    params = {name: value for name, value in (params or {}).items() if value is not None}

    async def aux():
        ssl_context = ssl.create_default_context(cafile=certifi.where())
        conn = TCPConnector(ssl=ssl_context)

        client = ClientSession(connector=conn, timeout=ClientTimeout(timeout))

        await asyncio.sleep(sleep)

        async with client as session:
            logger.info("REST: GET %s", url)

            async with session.get(url, headers=headers, params=params) as response:
                status_code = response.status

                try:
                    json_content = await response.json()
                except ContentTypeError:
                    json_content = None

            await response.release()

        if status_code == 429:
            if "Retry-After" in response.headers:  # opensea.io
                wait = int(response.headers["Retry-After"]) + 5
                logger.info("too many requests, waiting for %ss", wait)
                await asyncio.sleep(wait)
                response.raise_for_status()
            elif json_content.get("detail") == "Request was throttled.":  # opensea.io
                wait = 90
            else:
                wait = 30

            logger.info("too many requests, waiting for %ss", wait)
            await asyncio.sleep(wait)
            response.raise_for_status()

        if check:
            assert check in json_content

        return json_content, status_code

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    result = loop.run_until_complete(aux())
    loop.close()

    return result


@backoff.on_exception(backoff.expo, exception=Exception, max_time=60)
def post(url: str, body: dict = None, headers: dict = None, params: dict = None, sleep: float = 0.05, timeout: float = None) -> Tuple[dict, int]:
    headers = {name: value for name, value in (headers or {}).items() if value is not None}
    params = {name: value for name, value in (params or {}).items() if value is not None}

    async def aux():
        ssl_context = ssl.create_default_context(cafile=certifi.where())
        conn = TCPConnector(ssl=ssl_context)

        client = ClientSession(connector=conn, timeout=ClientTimeout(timeout))

        await asyncio.sleep(sleep)

        async with client as session:
            logger.info("REST: POST %s", url)

            async with session.post(url, json=body, headers=headers, params=params) as response:
                status_code = response.status

                try:
                    json_content = await response.json()
                except ContentTypeError:
                    json_content = None

                await response.release()

            if status_code == 429:
                if "Retry-After" in response.headers:  # opensea.io
                    wait = int(response.headers["Retry-After"]) + 5
                    logger.info("too many requests, waiting for %ss", wait)
                    await asyncio.sleep(wait)
                    response.raise_for_status()
                elif json_content.get("detail") == "Request was throttled.":  # opensea.io
                    wait = 90
                else:
                    wait = 30

                logger.info("too many requests, waiting for %ss", wait)
                await asyncio.sleep(wait)
                response.raise_for_status()

            return json_content, status_code

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    result = loop.run_until_complete(aux())
    loop.close()

    return result


@backoff.on_exception(backoff.expo, exception=Exception, max_time=60)
def get_stream(
    url: str, headers: dict = None, params: dict = None, sleep: float = 0.05, timeout: float = None, check: str = None
) -> Tuple[dict, int]:
    headers = {name: value for name, value in (headers or {}).items() if value is not None}
    params = {name: value for name, value in (params or {}).items() if value is not None}

    async def aux():
        ssl_context = ssl.create_default_context(cafile=certifi.where())
        conn = TCPConnector(ssl=ssl_context)

        client = ClientSession(connector=conn, timeout=ClientTimeout(timeout))

        await asyncio.sleep(sleep)

        async with client as session:
            logger.info("REST: STREAM %s", url)

            async with session.get(url, headers=headers, params=params) as response:
                status_code = response.status

                try:
                    content = []

                    async for chunk in response.content:
                        content.append(json.loads(chunk.decode("utf-8")))

                    json_content = content
                except ContentTypeError:
                    json_content = None

            await response.release()

        if status_code == 429:
            if "Retry-After" in response.headers:  # opensea.io
                wait = int(response.headers["Retry-After"]) + 5
                logger.info("too many requests, waiting for %ss", wait)
                await asyncio.sleep(wait)
                response.raise_for_status()
            elif json_content.get("detail") == "Request was throttled.":  # opensea.io
                wait = 90
            else:
                wait = 30

            logger.info("too many requests, waiting for %ss", wait)
            await asyncio.sleep(wait)
            response.raise_for_status()

        if check:
            assert json_content.get(check)

        return json_content, status_code

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    result = loop.run_until_complete(aux())
    loop.close()

    return result



from .asset import AssetDTO
from .floor import FloorDTO
from .holding import HoldingDTO
from .listing import ListingDTO
from .rank import RankDTO
from .sale import SaleDTO, SaleType
from .transaction import TransactionDTO, TransactionInputOutputDTO, TransactionAssetDTO



import logging
import os
from typing import Tuple

from fetchfox import rest
from fetchfox.checks import check_str
from fetchfox.constants.blockchains import ETHEREUM, POLYGON

BASE_URL_DEFAULT = "https://deep-index.moralis.io/api"
BASE_URL = os.getenv("MORALISIO_BASE_URL") or BASE_URL_DEFAULT

API_KEY = os.getenv("MORALISIO_API_KEY")

logger = logging.getLogger(__name__)


MAINNET = {
    ETHEREUM: "eth",
    POLYGON: "polygon",
}

TESTNET = {
    ETHEREUM: "0xaa36a7",  # sepolia
    POLYGON: "0x13882",  # amoy
}


def get(
    service: str,
    blockchain: str,
    params: dict = None,
    version: str = "2.2",
    api_key: str = None,
    preprod: bool = False,
) -> Tuple[dict, int]:
    api_key = api_key or API_KEY
    check_str(api_key, "moralisio.api_key")
    check_str(blockchain, "moralisio.blockchain")

    if preprod:
        chain = TESTNET.get(blockchain)
    else:
        chain = MAINNET.get(blockchain)

    params = params or {}
    params["chain"] = chain

    return rest.get(
        url=f"{BASE_URL}/v{version}/{service}",
        params=params,
        headers={
            "X-API-Key": api_key,
            "Host": "deep-index.moralis.io",
        },
    )
